comment *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
            siebneu.asm

            New and optimzed version of sieb1.asm
            
                            SUBSYSTEM:CONSOLE
        *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        
include     c:\masm32\include\masm32rt.inc
; 

DumpArray PROTO :DWORD  ; count and/or show/store the prime numbers found

.data?
    pMem        dd  ?   ; pointer to array of numbers[0..cByt]
    cByt        dd  ?   ; find primes up to this value
    primesfound dd  ?   ; number of primes between [0..cByt]
    
    ms_init     dd  ?   ; time needed to initialize array
    ms_sieve    dd  ?   ; time needed for sieving
    ms_count    dd  ?   ; .....for counting and/or saving primes
    cnt_loop    dd  ?   ; loop counter
    
.data
    startmsg    db  13,10,"Sieve of Eratosthenes - find primes up to: ",0
; 

.code
start:
    print addr startmsg
    mov cByt,uval(input())  ; user enters max_number (unsigned, because
                            ; there are no negative primes!  :-)
                            ; thanks to jj2007
    .if cByt==0             ; if user enters nothing or NULL
        mov cByt,500000     ; we take 500000 as default
    .endif
    add cByt,1  ; array starts with [0], so add 1 to cByt
                ; otherwise, we search only up to cByt - 1
        
    invoke GlobalAlloc,GMEM_FIXED,cByt  ; allocate cByt bytes of memory
    mov pMem,eax                        ; store array-address to pMem
    .if eax==NULL
        print "Not enough memory!"
        exit
    .endif
; -------------------------------------------------------------------------
    print "Looking for primes from 1 to "
    sub cByt,1
    print ustr$(cByt), " - please wait...",13,10,13,10
    add cByt,1
;---------------------------------------------------------------
; Start timing: initialize Array
;---------------------------------------------------------------
    invoke GetTickCount
    mov ms_init,eax
@@:
    invoke GetTickCount     ; wait until next tick
    .if ms_init==eax        ; (timing is more reliable, because
        jmp @B              ; we start _exactly_ at the beginning
    .endif                  ; of a new tick)
    mov ms_init,eax
;---------------------------------------------------------------
; initialize Array
;---------------------------------------------------------------
    mov esi,pMem
    mov dword ptr[esi+0],000000101h         ; preset 0,1,2
    push cByt                               ; save cByt on stack
    shr cByt,2                              ; Cbyt\4
    add cByt,1                              ; insure,we don't miss last 4 bytes
    mov edx,4                               ; we work with DWORDs
align 4
    .repeat
        mov dword ptr[esi+edx],000010001h   ; preset all even numbers = 1
        add edx,4
    .until edx >= cByt
    pop cByt                                ; get cByte back
    invoke GetTickCount
    sub eax,ms_init
    mov ms_init,eax
    print ustr$(ms_init)," ms to initialize array",13,10
;---------------------------------------------------------------
; Start timing: sieving
;---------------------------------------------------------------
    invoke GetTickCount
    mov ms_sieve,eax
@@:
    invoke GetTickCount
    .if ms_sieve==eax
        jmp @B
    .endif
    mov ms_sieve,eax
;---------------------------------------------------------------
; sieve primes from 0 .. cByt+1
;---------------------------------------------------------------
    mov cnt_loop,0
    mov ebx,3                               ; ebx = i we start with 3
    mov eax,3
    mul eax                                 ; eax = i*i
align 4
    .while eax < cByt                       ; while i*i < N
        .if byte ptr[esi+ebx]==0            ; if not deleted[i] then
            mov edx,eax                     ;    ebx = j
            align 4
            .while edx < cByt               ; for j = i*i to N
                .if (edx & 1)               ;   if j mod 2 = 0
                    inc cnt_loop            ;      increment loop-counter
                    mov byte ptr[esi+edx],1 ;      deleted[j] = true
                .endif
                add edx,ebx
            .endw
        .endif
        add ebx,2   ; now becomes 5,7,9 and so on, don't need 2,4,6
        mov eax,ebx
        mul eax                             ; eax = i*i
    .endw
;---------------------------------------------------------------
    invoke GetTickCount
    sub eax,ms_sieve
    mov ms_sieve,eax
    print ustr$(ms_sieve)," ms to sieve the primes",13,10
    
    invoke GetTickCount
    mov ms_count,eax
    
    invoke DumpArray,esi    ; process bytearray, inc primesfound for every
                            ; number that is not deleted

    print chr$(13,10)
    print ustr$(cnt_loop)," loop-counter",13,10
    
    invoke GlobalFree,pMem
    .if eax!=NULL
        push eax
        print chr$(13,10)
        pop eax
        print hex$(eax)
        print " - ERROR GlobalFree"
    .endif
    exit
;---------------------------------------------------------------
align 4
DumpArray proc pArr:DWORD

    LOCAL hFile:DWORD
    LOCAL rlen :DWORD
        
    mov esi,pArr
    mov ebx,3
    mov primesfound,1
    ;print "2",9

; Uncomment next block, if you want to store the primes to "primes.txt"
; -------------------------------------------------------------------------
;
;    invoke CreateFile,chr$("primes.txt"),
;                      GENERIC_READ or GENERIC_WRITE,
;                      NULL,
;                      NULL,
;                      CREATE_ALWAYS,
;                      FILE_ATTRIBUTE_NORMAL,
;                      NULL
;    mov hFile,eax
;    
; -------------------------------------------------------------------------

.data
    szPrim  db 10 dup(?),0
.code    
align 4
    .while ebx < cByt
        .if byte ptr[esi+ebx]==0
            ;--------------------------------------------------------
            ; uncomment following 2 lines to save result in "primes.txt"
            ;
            ;invoke dwtoa,ebx,addr szPrim   ; convert prime to ASCII
            ;invoke WriteFile,hFile,ADDR szPrim,10,ADDR rlen,NULL
            ;
            ;invoke WriteFile,hFile,ebx,4,ADDR rlen,NULL    ; write 'raw'
            ;
            ; or
            ;
            ;print ustr$(ebx),9
            ;
            ; for output in console            
            ;--------------------------------------------------------
            inc primesfound
        .endif
        add ebx,2
    .endw
    invoke GetTickCount
    sub eax,ms_count
    mov ms_count,eax
    print ustr$(ms_count)," ms to count the primes",13,10,13,10

    mov eax,ms_init
    add eax,ms_sieve
    add eax,ms_count
    print ustr$(eax)," ms total time."
    
    print chr$(13,10,13,10)
    sub cByt,1
    print "  Number of primes found between 1 and "
    print ustr$(cByt)," = "
    print ustr$(primesfound)
    
    ret
DumpArray endp
end start